/* slurp-manager.c
 *
 * Copyright © 2018 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "slurp-download.h"
#include "slurp-manager-job.h"
#include "slurp-manager.h"

typedef struct _SlurpManagerPrivate
{
    gint limit;

    GThreadPool *worker;
    SlurpDownload *download;
} SlurpManagerPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpManager, slurp_manager, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_LIMIT,
    NUM_PROPERTIES
};

static GParamSpec *properties[NUM_PROPERTIES];

enum {
    SLURP_MANAGER_ERROR,
    SLURP_MANAGER_ERROR_WORKER_TASK_INVALID,
    SLURP_MANAGER_ERROR_WORKER_TASK_CANCELLED
};

void
_manager_add_resource (SlurpManager  *self,
                       SlurpResource *resource,
                       GCancellable  *cancellable)
{
    SlurpManagerPrivate *priv;
    GTask *task;

    g_return_if_fail (SLURP_IS_MANAGER (self));
    g_return_if_fail (SLURP_IS_RESOURCE (resource));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    priv = slurp_manager_get_instance_private (self);

    task = g_task_new (self, cancellable, NULL, NULL);
    g_task_set_source_tag (task, _manager_add_resource);
    g_task_set_task_data (task, resource, (GDestroyNotify) g_object_unref);

    g_print ("# Added task\n");
    // TODO: Add error, kill task if it fails or retry
    g_thread_pool_push (priv->worker, task, NULL);
}

static void
_manager_download_cb (GObject      *object,
                      GAsyncResult *result,
                      gpointer      user_data)
{
    //SlurpDownload *download = SLURP_DOWNLOAD (object);
    GTask *task = G_TASK (user_data);
    //SlurpManager *self = SLURP_MANAGER (g_task_get_source_object (task));
    //SlurpManagerPrivate *priv = slurp_manager_get_instance_private (self);
    gboolean retval;
    GError *error;

    retval = slurp_download_resource_finish (object, result, &error);

    g_print ("STATUS: %s\n", retval ? "true" : "false");

    if (!retval)
    {
        g_debug ("Error: %s", error->message);
    }

    g_object_unref (task);
    g_print ("# Finished task\n");
}

static void
_manager_thread_worker_func (gpointer data,
                             gpointer user_data)
{
    GTask *task = G_TASK (data);
    SlurpManager *self = SLURP_MANAGER (user_data);
    SlurpManagerPrivate *priv = slurp_manager_get_instance_private (self);
    SlurpResource *resource;
    GCancellable *cancellable;

    g_return_if_fail (G_IS_TASK (task));
    g_return_if_fail (SLURP_IS_MANAGER (self));

    if (task == NULL)
    {
        return;
    }

    if (!g_task_is_valid (task, self))
    {
        g_task_return_new_error (task,
                                 SLURP_MANAGER_ERROR,
                                 SLURP_MANAGER_ERROR_WORKER_TASK_INVALID,
                                 "GTask is invalid, skipping download");

        g_object_unref (task);
        return;
    }

    if (g_task_return_error_if_cancelled (task))
    {
        g_task_return_new_error (task,
                                 SLURP_MANAGER_ERROR,
                                 SLURP_MANAGER_ERROR_WORKER_TASK_CANCELLED,
                                 "GTask is cancelled, skipping download");
        g_object_unref (task);
        return;
    }

    resource = SLURP_RESOURCE (g_task_get_task_data (task));
    cancellable = g_task_get_cancellable (task);

    slurp_download_resource_async (priv->download,
                                   resource,
                                   cancellable,
                                   _manager_download_cb,
                                   task);
}

static void
slurp_manager_constructed (GObject *object)
{
    SlurpManager *self = SLURP_MANAGER (object);
    SlurpManagerPrivate *priv = slurp_manager_get_instance_private (self);
    GError *error = NULL;

    priv->worker = g_thread_pool_new (_manager_thread_worker_func,
                                      self,
                                      priv->limit,
                                      FALSE,
                                      &error);

    if (priv->worker == NULL)
    {
        g_critical ("Unable to start slurp-worker thread");
        return;
    }

    priv->download = slurp_download_new ();
}

static void
slurp_manager_finalize (GObject *object)
{
    SlurpManager *self = SLURP_MANAGER (object);
    SlurpManagerPrivate *priv = slurp_manager_get_instance_private (self);

    g_clear_pointer (&priv->worker, g_object_unref);
    g_clear_pointer (&priv->download, g_object_unref);

    G_OBJECT_CLASS (slurp_manager_parent_class)->finalize (object);
}

static void
slurp_manager_get_property (GObject    *object,
                            guint       prop_id,
                            GValue     *value,
                            GParamSpec *pspec)
{
    SlurpManager *self = SLURP_MANAGER (object);
    SlurpManagerPrivate *priv = slurp_manager_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_LIMIT:
            g_value_set_int (value, priv->limit);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_manager_set_property (GObject      *object,
                            guint         prop_id,
                            const GValue *value,
                            GParamSpec   *pspec)
{
    SlurpManager *self = SLURP_MANAGER (object);
    SlurpManagerPrivate *priv = slurp_manager_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_LIMIT:
            priv->limit = g_value_get_int (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_manager_class_init (SlurpManagerClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_manager_finalize;
    object_class->constructed = slurp_manager_constructed;
    object_class->get_property = slurp_manager_get_property;
    object_class->set_property = slurp_manager_set_property;

	/**
	 * SlurpManager:limit:
	 *
	 * The number of allowed concurrent downloads.
	 */
    properties[PROP_LIMIT] =
        g_param_spec_int ("limit",
                          "Concurrent Download Limit",
                          "The number of allowed concurrent downloads.",
                          1,
                          G_MAXINT,
                          5,
                          G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       properties);
}

static void
slurp_manager_init (SlurpManager *self)
{
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
SlurpManager *
slurp_manager_new (void)
{
    return g_object_new (SLURP_TYPE_MANAGER, NULL);
}

void
slurp_manager_add_resource (SlurpManager  *self,
                            SlurpResource *resource,
                            GCancellable  *cancellable)
{
    g_return_if_fail (SLURP_IS_MANAGER (self));
    g_return_if_fail (SLURP_IS_RESOURCE (resource));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    _manager_add_resource (self, resource, cancellable);
}
